/*
 *  tn3270 for the Macintosh Source Code
 *  Brown University Computing and Information Services
 *  Version 2.3d21, January 17, 1991
 *  Copyright (c) 1988, 1989, 1990, 1991 by Brown University and by
 *  Peter John DiCamillo.
 *
 *  Permission is granted to any individual or institution to use, copy,
 *  or redistribute the binary version of this software and its
 *  documentation provided this notice and the copyright notices are
 *  retained.  Permission is granted to any individual or non-profit
 *  institution to use, copy, modify, or redistribute the source files
 *  of this software provided this notice and the copyright notices are
 *  retained.  This software may not be distributed for profit, either
 *  in original form or in derivative works, nor can the source be
 *  distributed to other than an individual or a non-profit institution.
 *  Any  individual or group interested in seeing and/or using these
 *  source files but who are prevented from doing so by the above
 *  constraints should contact Don Wolfe, Assistant Vice-President for
 *  Computer Systems at Brown University, (401) 863-7250, for possible
 *  software licensing of the source developed at Brown.
 *
 *  Brown University and Peter John DiCamillo make no representations
 *  about the suitability of this software for any purpose.
 *
 *  BROWN UNIVERSITY AND PETER JOHN DICAMILLO GIVE NO WARRANTY, EITHER
 *  EXPRESS OR IMPLIED, FOR THE PROGRAM AND/OR DOCUMENTATION PROVIDED,
 *  INCLUDING, WITHOUT LIMITATION, WARRANTY OF MERCHANTABILITY AND
 *  WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "maclib.h"
#include "termdef.h"
#include "globals.h"
#include "telnet.h"

#define adMin 		39
#define adMax 		37

/* location for disk initialization dialog */
#define kDITop		0x0050
#define kDILeft		0x0070

#define gestaltPPCSupportsRealTime	0x1000

extern char vmxbgn, vmxsub, kabort;
extern char xdlg;
extern DialogPtr xdlgptr;
extern Rect gr_rect;
extern short vnum;
extern char hisopts[];
extern vmprt prtinfo;
extern char skipnewmode;
extern Handle rcvMessage;			/* PPC message buffers */
extern Handle sndMessage;

struct Point sfppoint, sfgpoint;	/* standard file dialog locations */
char da_menu = 0;					/* TRUE when DA menu should be displayed */
char showstg = 0;					/* True when storage info. wanted */
int button;							/* button down at startup */
char tcpinitok = 0;					/* tcp network stuff active */
char pend_conn = 0;					/* pending connection state (TCP/IP) */
short systemvol;					/* volume with system files */
int defaultv;						/* default volume for TCP/IP FTP */

Rect statRect, textRect, scrnRect, pictRect;
Rect cposRect;						/* rectangle for cursor postion numbers */
GDHandle bitmap_gdev;				/* gdev used for bitmaps */
GDHandle saved_gdev;				/* gdev for set/rest gdev */

unsigned char ordertab[256];		/* table checking for 3270 order vs. data */
unsigned char nlbuff[80];			/* buffer for language translation */
									/* size = screen width */
Handle alaxtabh;					/* handle to ALA diacritic table */

short hpixsize, vpixsize, scrvoff, scrhsize, scrhoff;
short htxtstrt, vtxtstrt, vtxtsize, htxtsize, curvsize, curhsize;
short curnum, x_origin, y_origin;
short maxcnt, maxoff;
long textsize, pictsize;	/* size in bytes for stginfo */
short launchvol;			/* default volume at launch */
char newserver;				/* Telnet server supports new protocol */
char statcpos;				/* only draw curor position in status line */
char smallscreen;			/* complete 24 x 80 window doesn't fit */
char pendalign;				/* alignbitmap call pending */
char serverconn = 0;		/* connecting to non-terminal server*/
char servermode = 0;		/* server connection active */
char serverflags = 0;		/* flags bits for special server processing */
char notifyavail = 0;		/* Notification Manager available */
nmrmessage *myNMR[NMRNUM];	/* pointers to Notification Manager messages */

PaletteHandle myPalette;
short pxsize;
RGBColor realblack = {0, 0, 0};
RGBColor realwhite = {0xffff, 0xffff, 0xffff};

Ptr myNewPtr();

main()
{
short rc;
char dummy[28];
short rid;
ResType rtype;
Rect deskrect;
unsigned short height, width;
Handle temph;
short i;

int_active = wr_active = 0;
aplmode = insmode = online = 0; /* init. status line */
drawpict = 0;
newpict = 1;
savepict = 0;			/* not saving graphics in PICT file */
textsel = 0;			/* no text selected */
pndinv = pndclr = pndnewmode = 0;
needwrite = 0;
kb_err = 0;
kbqsize = 0;
ioerror = 0;
logon = 0;
kblock = 1;
kblcode = 4;
ftpcopen = ftpdopen = ftplopen = 0;		/* no TCP/IP FTP yet */
kb_handle = 0;			/* no keyboard mapping yet */
nl_handle = 0;			/* no language mapping yet */
stdfont = NORMALFONT;	/* use normal 3278 font */
plainala = 0;			/* overstrike with ALA */
sndhandle = 0;			/* default (system) sound */
sndactive = 0;			/* no sound currently active */
prtinit = 0;
myWindow = 0;			/* no window yet */
atrbuff = 0;			/* no attribute buffer */
readbuff = 0;			/* no read-modified response buffer */
chrbuff = 0;			/* no 3270 screen buffer */
scrapstart = 0;			/* no private scrap storage */
mapptr = pmapptr = 0;	/* no bitmaps yet */
textmap = 0;			/* no text bitmap yet */
x = y = 0;				/* initial cursor position */
newserver = 0;			/* default 0; only set by TCP code */
statcpos = 0;			/* draw cursor position only */
smallscreen = 0;		/* assume window will fit */
pendalign = 0;			/* alignbitmap call not pending */
skipnewmode = 0;		/* call newmode for EW/EWA as usual */
wposok = 0;				/* window position not set yet */
smgr_sppc = sess_sppc = 0;	/* no SPPC registrations yet */
smgrppc = sessppc = 0;	/* no PPC ports yet */
rcvMessage = sndMessage = 0;	/* no ppc buffers yet */
GetVol(ordertab, &launchvol);	/* define launchvol */
apiopen = 0;			/* apiopenreq not in use */
apiclosenotify = 0;		/* no close notification */

						/* define handle to EBCDIC to ASCII table */
xtabh = GetResource('GFXT', 128);
alaxtabh = GetResource('GFXT', 129);
						/* define prtinfo */
temph = GetResource('VMPR', 128);
if (temph != 0) {
	movmem(*temph, &prtinfo, sizeof(vmprt));
	if (prtinfo.linewidth > MAXLINEWIDTH) prtinfo.linewidth = MAXLINEWIDTH;
	if (prtinfo.titlewidth > MAXLINEWIDTH) prtinfo.titlewidth = MAXLINEWIDTH;
	}
	
orderinit();			/* initialize ordertab */

macinit();				/* general Mac initialization */
hnd_init();				/* handle any update events */

						/* calculate limits for screen format */
getdeskrect(&deskrect);
height = deskrect.bottom;
height -= deskrect.top;
height -= 47;
		/* If the top of the desk rectangle is <= 0, it includes the
		   menu bar, and the size of the menu bar is subtracted */
if (deskrect.top <= 0) {
	height -= 20;
	}
width = deskrect.right;
width -= deskrect.left;
width -= 13;

	/* calculate 9 and 12-point limits */
rowmax9 = height/12;
if (rowmax9 < 24) {
	rowmax9 = 24;
	smallscreen = 1;
	}
colmax9 = width/6;
if (colmax9 < 80) {
	colmax9 = 80;
	}
rowmax12 = height/16;
colmax12 = width/7;

rc = init_settings();	/* get initial settings */

						/* set keymap handle if not already set */
if (kb_handle == 0L) {
	kb_handle = GetNamedResource('GFKB', cskybdname);
	if (kb_handle == 0L) {
		stoperr(kbdefalrt);
		ExitToShell();
		}
	else {
		GetResInfo(kb_handle, &rid, &rtype, fskybdname);
		strcpy(cskybdname, fskybdname);
		}
	}
					/* initialize for connection type */
if (serflg) {			/* serial */
	ser_init();
	}
else if (tcpflg) {		/* tcpip */
	tcp_init();
	if (ioerror != 0) {
		macend();
		return;
		}
	}
else {					/* AppleTalk */
	net_init();
	}

if (rc != 0) {			/* set default host if no settings file */
	if (serflg) serdflthost(cshostname);
	else if (tcpflg) tcpdflthost(cshostname);
	else netdflthost(cshostname);
	strcpy(fshostname, cshostname);
	}

GetVol(dummy, &vnum);	/* init. default volume for file transfer */
defaultv = vnum;		/* also set for FTP */
xfinit();
				/* login if net connection, settings file, and autoconn */
if ((!serflg) && (rc == 0) && cs.autoconn) {
	if (tcpflg) tcplgin();
	else netlgin();
	}

done = false;
while (!done) hndmac();		/* handle Mac events and I/O */
 
if (serflg) ser_end();		/* clean-up I/O */
else if (tcpflg) tcp_end();
	 else net_end();
macend();					/* clean-up Mac */
}

macinit()
{
short i;
static char * ScrDmpEnb = (char *)0x2f8;
pascal void attnmenu();
static void (*attnhdl) ();
SysEnvRec theWorld;
GDHandle currGD;
PixMapHandle currPM;
char *s;
Rect deskrect;
OSErr rc;

			/* set-up general Macintosh environment */
MaxApplZone();				/* set-up for efficient storage use */
for (i=0; i < 4; i++) MoreMasters();
InitGraf(&qd.thePort);
FlushEvents(everyEvent, 0);
InitWindows();
InitFonts();
InitMenus();
InitDialogs(0L);
InitCursor();
TEInit();

			/* set-up menus */
for (i=0; i < NMENUS; i++) {
	myMenus[i] = GetMenu(256+i);
	if (i == 4) {		/* attn key menu */
		attnhdl = attnmenu;
		(*myMenus[4])->menuProc = (Handle)&attnhdl;
		CalcMenuSize(myMenus[4]);
		(*myMenus[4])->enableFlags = 1;
		}
	if (i < 5) InsertMenu(myMenus[i], 0);
		/* 5th menu inserted by tcp_init */
	}
AddResMenu(myMenus[0], 'DRVR');

kybdMenu = GetMenu(235);
InsertMenu(kybdMenu, -1);
AddResMenu(kybdMenu, 'GFKB');

nlMenu = GetMenu(234);
InsertMenu(nlMenu, -1);
AddResMenu(nlMenu, 'GFNL');

/* define csnlname from the name of the first item in the */
/* national language menu */
GetItem(nlMenu, 1, csnlname);

appl_menu();
DrawMenuBar();

			/* check if mouse is down at startup */
button = Button(); 

			/* get environment information */
SysEnvirons(1, &theWorld);		
newroms = (theWorld.machineType >= 0);
colormac = (theWorld.hasColorQD != 0);
systemvol = theWorld.sysVRefNum;
syslevel = theWorld.systemVersion >> 8;
notifyavail = (syslevel >= 6);

if (colormac) {		/* We have color QuickDraw, now check if		*/
					/* the deepest display device supports color.	*/
					/* We won't use color if the device is			*/
					/* monochrome and the current pixel size		*/
					/* is 1.  Ideally, we would be able to			*/
					/* decide based on the maximum pixel size.		*/
	getdeskrect(&deskrect);
	currGD = GetMaxDevice(&deskrect);
	currPM = (*currGD)->gdPMap;
	if ((*currPM)->pixelSize == 1) 
		colormac = (*currGD)->gdFlags & (1 << gdDevType);
	}

/* determine if Gestalt is available */
gestaltavail = 0;
if (newroms) {	   /* compare Gestalt OS trap to Unimplemented Toolbox trap */
	gestaltavail = NGetTrapAddress(0xAD, 0) !=
				   NGetTrapAddress(0x9F, 1);
	}

/* determine availability of PPC */
ppcavail = 0;
ppcflags = 0;
if (gestaltavail) {
	rc = Gestalt('ppc ', &ppcflags);
	ppcavail = (rc == noErr);
	if (!ppcavail) ppcflags = 0;
	}

screenRect = qd.screenBits.bounds;

			/* read alerts into memory and center */
ctralrt(263);
ctralrt(264);
ctralrt(265);
ctralrt(274);
ctralrt(275);
ctralrt(276);
ctralrt(278);
ctralrt(279);
ctralrt(281);

			/* calculate standard file dialog location*/
sfppoint.h = (screenRect.right-screenRect.left-304)/2;
sfgpoint.h = (screenRect.right-screenRect.left-348)/2;
if (smallscreen) {
	sfppoint.v = (screenRect.bottom-screenRect.top-184)/2;
	sfgpoint.v = (screenRect.bottom-screenRect.top-200)/2;
	}
else {
	sfppoint.v = (screenRect.bottom-screenRect.top-184)/3;
	sfgpoint.v = (screenRect.bottom-screenRect.top-200)/3;
	}

if (colormac) {
	getdeskrect(&dragRect);
	}
else {
	SetRect(&dragRect, 0, 24, screenRect.right-4, screenRect.bottom-4);
	}

			/* allocate storage for "fs" string variables and api
			   variables :
			   fshostname = 128
			   fskybdname = 256
			   fsnlname = 256
			   fssndname = 256
			   apihostname = 256
			   apiwindowname = 128  total = 1280 bytes */


s = NewPtr((Size)1280);
if (s == 0L) {
	stgalert("fs_and_api_variables", "NewPtr", (Size)1280);
	ExitToShell();
	}
fshostname = s;
fskybdname = s + 128;
fsnlname = s + 384;
fssndname = s + 640;
apihostname = s + 896;
apiwindowname = s + 1152;

			/* allocate storage for notification manager */
for (i=0; i < NMRNUM; i++) {
	myNMR[i] = (nmrmessage *)NewPtr((long)(sizeof(nmrmessage)));
	if (myNMR[i] == 0L) {
		notifyavail = 0;
		stgalert("nmrmessages", "NewPtr", (long)(sizeof(nmrmessage)));
		ExitToShell();
		}
	else {
		memset(myNMR[i], 0, sizeof(nmrmessage));		
		}
	}

			/* allocate sound channel storage (about 1K) */
scp = (SndChannelPtr)NewPtr((long)(sizeof(struct SndChannel)));
if (scp == 0L) {
	stgalert("sound channel", "NewPtr", (long)(sizeof(struct SndChannel)));
	ExitToShell();
	}
setmem(scp, sizeof(struct SndChannel), 0);
scp->qLength = stdQLength;


			/* allocate print record */
hPrint = (THPrint)NewHandle((long)sizeof(struct TPrint));
if (hPrint == 0L) {
	stgalert("print record", "NewHandle", (long)(sizeof(struct TPrint)));
	ExitToShell();
	}

			/* allocate work regions for grow icon */
cliprgn = NewRgn();	
destrgn = NewRgn();
temprgn = NewRgn();
if ((cliprgn == 0L) || (destrgn == 0L) || (temprgn == 0L)) {
	stgalert("grow icon regions", "NewRgn", 0L);
	ExitToShell();
	}
	
			/* miscellaneous initialization */
CouldDialog(256);
ScrDmpSav = *ScrDmpEnb;		/* save current byte */
sw_bgrnd = 0;				/* not in switcher background */
mf_bgrnd = 0;				/* not in MultiFinder background */

/* if PPC is available, initialize it */
if (ppcavail) {
	if ((ppcflags & gestaltPPCSupportsRealTime) == 0) {
		rc = PPCInit();
		if (rc == noErr) {
			rc = Gestalt('ppc ', &ppcflags);
			}
		if (rc != noErr) {
			ppcavail = 0;
			ppcflags = 0;
			stoperr(ppcinitalrt);
			}
		}
	}

/* if ppc will be used, initialize queues */
if (ppcavail) {
	initparamblocks();
	initclientrecs();
	initdatablocks();
	addparamblocks();
	addclientrecs();
	adddatablocks();
	}

apiregister(0);		/* attempt to register session mgr with sppc and ppc */

}

hnd_init()				/* handle update events during intialization */
{
short evtrc;

evtrc = GetNextEvent(updateMask, &myEvent);
while (evtrc != 0) {
	switch(myEvent.what) {
		case updateEvt:
								updevent(myEvent.message);
								break;
		default:
								break;
		}
	evtrc = GetNextEvent(updateMask, &myEvent);
	}
}

/*	createwindow is called to create a new window when there is no
	current session (and no window).  At this point, the user's requested
	screen size will be used for the new window, and to determine the
	size of our private scrap. */

OSErr createwindow()
{
OSErr rc, newmode();
Size allocsize;

						/* set variables for this session */
if (apiopen) {
	ses_altrows = apiopenreq.altsize.rows;
	ses_altcols = apiopenreq.altsize.cols;
	ses_altptsize = apiopenreq.altsize.ptsize;
	ses_dfltptsize = apiopenreq.dfltsize.ptsize;
	}
else {
	ses_altrows = cs.altrows;
	ses_altcols = cs.altcols;
	ses_altptsize = cs.altptsize;
	ses_dfltptsize = cs.dfltptsize;
	}
ses_windmax = cs.windmax;
addr14 = (ses_altrows * ses_altcols) > 4096;

						/* allocate scrap of proper size */
rc = newscrap();
if (rc != noErr) {
	stoperr(newwindalrt);
	return(rc);
	}
/* In following code, no need to clean up scrap, since scrap persists,
   in order to not lose its contents needlessly.  (Newscrap will keep the
   existing scrap if it has the proper line width for this session.)	*/

						/* allocate 3270 screen buffers */
							/* 3270 attribute buffer */
allocsize = ses_altrows * ses_altcols;
allocsize <<= 1;
atrbuff = (short *)myNewPtr(allocsize);
if (atrbuff == 0L) {
	stoperr(newwindalrt);
	return(2);
	}
setmem(atrbuff, (short)allocsize, 0);
							/* buffer for read responses */
	/* size based on vmxfer maximum response: 
	   6 byte prefix
	   2 bytes: DB
	   4608 bytes: 2304 bytes of data with IAC quoting
	   5 bytes: CRC suffix
	   2 bytes: IAC, EOR
	   1 byte padding
	   ==> 4624 bytes		*/

readbuff = (unsigned char *)myNewPtr((long)4624L);	
if (readbuff == 0L) {
	DisposPtr(atrbuff);
	stoperr(newwindalrt);
	return(3);
	}
setmem(readbuff, 4624, 0);
rballoc = 4101;		/* limits responses to 4K (first 5 bytes not included) */

							/* 3270 screen image buffer */
allocsize = ses_altrows * ses_altcols;
chrbuff = (unsigned char *)myNewPtr(allocsize);
if (chrbuff == 0L) {
	DisposPtr(readbuff);
	DisposPtr(atrbuff);
	stoperr(newwindalrt);
	return(4);
	}
setmem(chrbuff, (short)allocsize, 0);
						/* create new window with alt. size */
ewamode = 1;
rc = newmode(1, 1);
if (rc != noErr) {
	DisposPtr(chrbuff);
	DisposPtr(readbuff);
	DisposPtr(atrbuff);
	stoperr(newwindalrt);
	return(rc);
	}
clrscn();
invldscr();
return(noErr);
}

removewindow()			/* completely remove window and buffers */
{						/* (returning to state with no window active) */
removemywindow();		
DisposPtr(chrbuff);
DisposPtr(readbuff);
DisposPtr(atrbuff);
}
				
removemywindow()		/* remove window in preparation for new mode */
{
updwindpos();
removebitmaps();
if (colormac) {
	restorectab();
	}
CloseWindow(myWindow);
myWindow = 0;
arrowcursor();
}

removebitmaps()
{
if (colormac) {
	if (WritePtr != myWindow) rmPix(WritePtr, mapptr);
	if (PictPtr != 0) rmPix(PictPtr, pmapptr);
	}
else {
	if (WritePtr != myWindow) ClosePort(WritePtr);
	if (mapptr != 0) DisposPtr(mapptr);
	if (PictPtr != 0) ClosePort(PictPtr);
	if (pmapptr != 0) DisposPtr(pmapptr);
	}
PictPtr = 0;
textmap = 0;
mapptr = pmapptr = 0;
textsize = pictsize = 0;
}

restorectab()
{
static unsigned char * RomMapInsert = (unsigned char *)0xb9e;
static unsigned char * TmpResLoad = (unsigned char *)0xb9f;
PaletteHandle oldpalette, fixpalette;
short pixdepth, colorcnt;
GDHandle currGD;
PixMapHandle currPM;
CTabHandle dflttab;

					/* attempt to restore default color table */
if (!colormac) return;		/* done if no color */
if (myWindow == 0) return;	/* return if no window */
if (syslevel > 6) return;	/* return if system handles color without our help */

currGD = GetGDevice();		/* get pixel depth */
currPM = (*currGD)->gdPMap;
pixdepth = (*currPM)->pixelSize;

colorcnt = 2 << (pixdepth-1);		/* compute number of colors */

							/* get default color table for pixel depth */
(*RomMapInsert) = 0xff;			/* include ROM in search */
(*TmpResLoad) = 0;				/* shouldn't need to load into memory */
dflttab = (CTabHandle)GetResource('clut', pixdepth);
if (dflttab == 0) return;		/* give up if no resource */

							/* use palette to set colors */
fixpalette = NewPalette(colorcnt+1, dflttab, pmTolerant, 0);
if (fixpalette == 0) return;			/* give up if can't get a palette */
oldpalette = GetPalette(myWindow);
SetPalette(myWindow, fixpalette, 0);	/* use new palette */
ActivatePalette(myWindow);
if (oldpalette != 0) {
	DisposePalette(oldpalette);			/* free previous palette */
	}
}
 
OSErr newmode(mode, create)
char mode;			/* 0 = normal; 1 = alternate */
char create;		/* 0 = change size; 1 = create window */
{			
short wproc;
BitMap PictMap;
BitMap WriteMap;
Size mapsize, pmapsize;
Ptr temp;
WStateData * wsptr;
short i, logrows, logcols;
short old_hpix, old_vpix;
Point p, q;
Rect r;
RGBColor rgbtemp;
PixMapHandle currPM;
short adj;
char cposstr[16];
OSErr alignbitmaps();
Boolean wvis;

/* calculate buffer size variables (logical rows and columns) */
if (mode == 0) {
	logrows = 24;
	logcols = 80;
	}
else {
	logrows = ses_altrows;
	logcols = ses_altcols;
	}

/* update the 3270 buffer variables */
maxcnt = logrows * logcols;
maxoff = maxcnt - 1;
scrvoff = logrows - 1;
scrhsize = logcols;
scrhoff = scrhsize - 1;
if (curadr > maxoff) curadr = maxoff;	/* keep cursor within text area */

/* handle AppleTalk call at interrupt level */
if (int_active) {
	pndmode = mode;
	pndnewmode = 1;
	return;
	}
else {
	pndnewmode = 0;
	}

/* calculate window size variables */
cur_mode = mode;
if (cur_mode == 1) {
	cur_rows = ses_altrows;
	cur_cols = ses_altcols;
	cur_ptsize = ses_altptsize;
	}
else {
	cur_rows = 24;
	cur_cols = 80;
	cur_ptsize = ses_dfltptsize;
	}
	
							/* save current window dimensions */
old_hpix = hpixsize;
old_vpix = vpixsize;
							/* set constants for new screen */
setmode(cur_rows, cur_cols, cur_ptsize, ses_windmax);

/* return if the current window is already the proper size */
if (myWindow != 0) {
	if ((old_hpix == hpixsize) && (old_vpix == vpixsize)) {
		return(0);
		}
	}

if (myWindow != 0) {				/* clean-up old window and bitmaps */
	if (create) removemywindow();	/* keep 3270 buffers */
	else removebitmaps();
	}

							/* keep cursor on screen */
if (curadr > maxoff) curadr = maxoff;

							/* change cursor if white needed */
if ((!cs.invertbw) && colormac) curnum += 2;

defcursor(0,0);				/* define er rectangle for screen cursor */

							/* define window rectangle */
if (smallscreen) {
	pRect.top = (screenRect.bottom-screenRect.top-vpixsize)/2 + 9;
	}
else {
	pRect.top = (screenRect.bottom-screenRect.top-vpixsize)/2 + 19;
	}
if ((pRect.top < 38) && (!smallscreen)) {
	pRect.top = 38;
	}
pRect.left = (screenRect.right-screenRect.left-hpixsize)/2 - 1;
if (pRect.left < 1) {
	pRect.left = 1;
	}
pRect.bottom = pRect.top+vpixsize;
pRect.right = pRect.left+hpixsize;

sizeRect.top = 25;
sizeRect.left = 50;
sizeRect.bottom = pRect.bottom - pRect.top + 1;
sizeRect.right = pRect.right - pRect.left + 1;

wproc = documentProc + 8;

if (myWindow == 0) {
	r = pRect;
	if (wposok) {	/* use stored window position if available */
		p.h = cs.windpth;
		p.v = cs.windptv;
		r.left = p.h - hpixsize/2 - 1;
		r.right = r.left + hpixsize + 3;
		r.top = p.v - vpixsize/2 - 10;
		r.bottom = r.top + vpixsize + 21;
		adjmove(&r);
		r.left += 1;			/* window rectangle -> port rectangle */
		r.right -= 2;
		r.top += 19;
		r.bottom -= 2;
		}
	if (colormac) {
		if (apiopen) {
			wvis = true;
			if (apiopenreq.hide != 0) wvis = false;
			myWindow = NewCWindow(&wRecord, &r, apiwindowname,
						wvis, wproc, -1L, true, 0L);
			}
		else {
			myWindow = NewCWindow(&wRecord, &r, wtitle,
						true, wproc, -1L, true, 0L);
			}
		if (myWindow == 0L) {
			return(5);
			}
		myPalette = NewPalette(11, 0L, pmCourteous, 0);
		if (myPalette == 0L) {
			CloseWindow(myWindow);
			return(6);
			}
		SetPalette(myWindow, myPalette, true);
		fixcolors(myPalette);		/* define colors to use */
		ActivatePalette(myWindow);
		}
	else {
		if (apiopen) {
			wvis = true;
			if (apiopenreq.hide != 0) wvis = false;
			myWindow = NewWindow(&wRecord, &r, apiwindowname,
						wvis, wproc, -1L, true, 0L);
			}
		else {
			myWindow = NewWindow(&wRecord, &r, wtitle,
						true, wproc, -1L, true, 0L);
			}
		if (myWindow == 0L) {
			return(7);
			}
		}
	}
else {
	SetPort(myWindow);
	/* change window size, and keep center of window constant */
	q.h = myWindow->portRect.left;	/* save old top left of portRect */
	q.v = myWindow->portRect.top;
	LocalToGlobal(&q);
	p.h = (myWindow->portRect.right + myWindow->portRect.left + 1)/2;
	p.v = (myWindow->portRect.bottom + myWindow->portRect.top - 17)/2;
	LocalToGlobal(&p);		/* global current center */
	r.left = p.h - hpixsize/2 - 1;
	r.right = r.left + hpixsize + 3;
	r.top = p.v - vpixsize/2 - 10;
	r.bottom = r.top + vpixsize + 21;
	adjmove(&r);
	r.left += 1;			/* window rectangle -> port rectangle */
	r.right -= 2;
	r.top += 19;
	r.bottom -= 2;
	p.h = r.left;
	p.v = r.top;
	if (((p.h - q.h) < 0) || ((p.v - q.v) < 0)) {	/* small -> large */
		MoveWindow(myWindow, p.h, p.v, true);
		SizeWindow(myWindow, hpixsize, vpixsize, true);
		}
	else {							/* large -> small */
		SizeWindow(myWindow, hpixsize, vpixsize, true);
		MoveWindow(myWindow, p.h, p.v, true);
		}
	InvalRect(&gr_rect);
	InitClip();
	InvalRect(&gr_rect);

	/* restore settings before bitmap allocation */
	if (colormac) {
		RGBBackColor(&realwhite);
		ActivatePalette(myWindow);
		}
	}

if (((WindowPeek)myWindow)->dataHandle != 0) {
	wsptr = (struct WStateData *)(*(((WindowPeek)myWindow)->dataHandle));
	wsptr->stdState = pRect;
	if (r == pRect) {
		adj = (pRect.bottom - pRect.top) / 4;
		wsptr->userState.top = pRect.top + adj;
		wsptr->userState.bottom = pRect.bottom - adj;
		adj = (pRect.right - pRect.left) / 4;
		wsptr->userState.left = pRect.left + adj;
		wsptr->userState.right = pRect.right - adj;
		}
	else {
		wsptr->userState = r;
		}
	zoomok = 1;
	}
else zoomok = 0;

SetPort(myWindow);
TextFont(stdfont);
TextSize(cur_ptsize);
x_origin = -4;
y_origin = -6;
SetOrigin(x_origin, y_origin);
InitClip();
grafcursor = GetCursor(curnum);
InitCursor();
dfltcurs = 1;	

			/* if color available, save pixel size for this window */
if (colormac) {
	GetGlobalRect(myWindow, &r);
	bitmap_gdev = GetMaxDevice(&r);
	currPM = (*bitmap_gdev)->gdPMap;
	pxsize = (*currPM)->pixelSize;
	}

textRect.top = scrnRect.top = y_origin;
statRect.left = textRect.left = scrnRect.left = x_origin;
statRect.top = textRect.bottom = y_origin + vpixsize - 19;
statRect.bottom = scrnRect.bottom = statRect.top + 20;
statRect.right = textRect.right = scrnRect.right = textRect.left + hpixsize;

				/* attempt to allocate text bitmap */
WriteMap.bounds.top = 0;
WriteMap.bounds.left = 0;
WriteMap.bounds.bottom = vpixsize;
WriteMap.bounds.right = hpixsize;
if (colormac) {
	mkPix(&WritePort, &mapptr, &(WriteMap.bounds), 0, bitmap_gdev);
	textsize = pictsize;
	pictsize = 0;
	if (mapptr == 0) {
		WritePtr = myWindow;
		note_err(mapalrt);
		}
	else {
		WritePtr = &WritePort;
		textmap = 1;
		}
	SetPort(WritePtr);
	GetEntryColor(myPalette, RGBback, &rgbtemp);
	if (WritePtr != myWindow) setgdev();
	RGBBackColor(&rgbtemp);
	if (WritePtr != myWindow) resetgdev();
	}
else {
	WriteMap.rowBytes = (((hpixsize + 31) >> 5) + 1) << 2;
	mapsize = WriteMap.bounds.bottom - WriteMap.bounds.top;
	mapsize = mapsize * WriteMap.rowBytes + 4;
	textsize = mapsize;
	mapptr = myNewPtr(mapsize);
	temp = mapptr;
	while (((Size)temp%4) != 0) temp++;	/* align on long word */
	WriteMap.baseAddr = temp;
	if (WriteMap.baseAddr == 0L) {		/* no screen buffer */
		WritePtr = myWindow;
		note_err(mapalrt);
		}
	else {
		WritePtr = &WritePort;			/* set-up WritePort */
		textmap = 1;
		WritePort.portRect = WriteMap.bounds;
		OpenPort(WritePtr);
		ClipRect(&WritePort.portRect);
		SetPortBits(&WriteMap);
		EraseRgn(WritePort.visRgn);
		TextFont(stdfont);
		TextSize(cur_ptsize);
		}
	}
if (WritePtr != myWindow) {
	SetPort(WritePtr);
 	SetOrigin(textRect.left, textRect.top);
	ClipRect(&scrnRect);
	SetPort(myWindow);
	}
 								/* set-up graphics bit map */
pictRect.top = pictRect.left = 0;
pictRect.bottom = vpixsize-19;
pictRect.right = hpixsize;

PictPtr = &PictPort;
if (!colormac) {
	PictMap.rowBytes = (((hpixsize + 31) >> 5) + 1) << 2;
	PictMap.bounds = pictRect;
	pmapsize = PictMap.bounds.bottom - PictMap.bounds.top;
	pmapsize = pmapsize * PictMap.rowBytes + 4;
	pictsize = pmapsize;
	pmapptr = myNewPtr(pmapsize);
	temp = pmapptr;
	while (((Size)temp%4) != 0) temp++;	/* align on long word */
	PictMap.baseAddr = temp;
	if (PictMap.baseAddr == 0L) {		/* no graphics bitmap */
		PictPtr = 0;
		}
	else {
									/* set-up PictPort */
		OpenPort(PictPtr);
		SetPortBits(&PictMap);
		PortSize(PictMap.bounds.right, PictMap.bounds.bottom);
		SetRectRgn(PictPort.visRgn, 0, 0, PictMap.bounds.right,
					PictMap.bounds.bottom);
		ClipRect(&PictPort.portRect);
		EraseRect(&PictPort.portRect);
		SetOrigin(0, 0);
		TextFont(stdfont);
		TextSize(cur_ptsize);
		}
	}
else {
	mkPix(&PictPort, &pmapptr, &pictRect, 0, bitmap_gdev);
	if (pmapptr == 0) PictPtr = 0;
	else {
		PictPtr = &PictPort;
		GetEntryColor(myPalette, RGBback, &rgbtemp);
		setgdev();
		RGBBackColor(&rgbtemp);
		resetgdev();
		}
	}

/* define status line cursor postion rectangle */
cposRect.top = statRect.top;
cposRect.bottom = statRect.bottom;
cposRect.left = 315;
sprintf(cposstr, "%02d/%02d", scrvoff+1, scrhsize);
TextFont(1);
TextSize(0);
cposRect.right = cposRect.left + StringWidth(cposstr);
TextFont(stdfont);
TextSize(cur_ptsize);
			/* for small screens, adjust cursor position between insert
			   mode symbol and time */
adj = 10 - hpixsize + 134 + cposRect.right;
if (adj > 0) {
	cposRect.left -= adj;
	cposRect.right -= adj;
	}

alignbitmaps();		/* always safe to call */
invldscr();
newstat();
return(0);
}

OSErr alignbitmaps()		
					/* Adjust bounds of our bitmaps for long word alignment
					   with the current window position.  Note that our
					   CopyBits calls copy the leftmost bit of each bitmap
					   to the leftmost bit of the window. */
					/* Non-zero return code indicates no need to redraw
					   text bitmap */
{
short pixeldepth;
GDHandle currGD;
PixMapHandle currPM;
GrafPtr gp;
Point p;
Rect r;
short adj;
BitMap * mapbits;

if (drawpict) {
	pendalign = 1;
	return(1);
	}
else {
	pendalign = 0;
	}
						/* nothing to do if no bitmaps */
if ((!textmap) && (PictPtr == 0)) {
	return(2);
	}

if (colormac) {
						/* if pixel depth has changed, return, since
						   we will be called for the new bitmap		*/
	GetGlobalRect(myWindow, &r);
	currGD = GetMaxDevice(&r);
	currPM = (*currGD)->gdPMap;
	if (pxsize != (*currPM)->pixelSize) return(3);
	if (pxsize > 16) return(4);		/* no adjustment needed */
	pixeldepth = pxsize;
	}
else {
	pixeldepth = 1;
	}

p.h = textRect.left;
p.v = textRect.top;
GetPort(&gp);
SetPort(myWindow);
LocalToGlobal(&p);
SetPort(gp);
getdeskrect(&r);
p.h = p.h - r.left;		/* ensure a positive value */
adj = p.h % (32 / pixeldepth);
if (PictPtr != 0) {
	if (colormac) {
		mapbits = (BitMap *)(*(PixMapHandle)PictPort.portBits.baseAddr);
		}
	else {
		mapbits = &PictPort.portBits;
		}
	(mapbits->bounds).left = -adj;
	}
if (textmap) {
	if (colormac) {
		mapbits = (BitMap *)(*(PixMapHandle)WritePort.portBits.baseAddr);
		}
	else {
		mapbits = &WritePort.portBits;
		}
	(mapbits->bounds).left = textRect.left - adj;
	return(0);
	}
else {
	return(5);
	}
}

setmode(dsprows, dspcols, ptsize, wmax)
short dsprows, dspcols, ptsize, wmax;
{
short dflthsize, dfltvsize, althsize, altvsize;
								/* set constants for desired format */
switch(ptsize) {
	case 12:
				htxtstrt = 0;
				vtxtstrt = 11;
				vtxtsize = 16;
				htxtsize = 7;
				curvsize = 16;
				curhsize = 7;
				curnum = 20846;
				break;
	default:
				htxtstrt = 1;
				vtxtstrt = 9;
				vtxtsize = 12;
				htxtsize = 6;
				curvsize = 12;
				curhsize = 7;
				curnum = 20845;
				break;
	}
if (wmax) {
	if (ses_dfltptsize == 9) {
		dflthsize = 80 * 6 + 10;
		dfltvsize = 24 * 12 + 26;
		}
	else {
		dflthsize = 80 * 7 + 10;
		dfltvsize = 24 * 16 + 26;
		}
	if (ses_altptsize == 9) {
		althsize = ses_altcols * 6 + 10;
		altvsize = ses_altrows * 12 + 26;
		}
	else {
		althsize = ses_altcols * 7 + 10;
		altvsize = ses_altrows * 16 + 26;
		}
	if (dflthsize > althsize) {
		hpixsize = dflthsize;
		}
	else {
		hpixsize = althsize;
		}
	if (dfltvsize > altvsize) {
		vpixsize = dfltvsize;
		}
	else {
		vpixsize = altvsize;
		}
	}
else {
	hpixsize = dspcols * htxtsize + 10;
	vpixsize = dsprows * vtxtsize + 26;
	}
}

newdepth()
{			
			/* re-allocate bitmaps for new pixel depth */
			/* (Macs with color QuickDraw only) 	   */
PixMapHandle currPM;
Rect r;
BitMap WriteMap;
RGBColor rgbtemp;
GrafPtr gp;
OSErr alignbitmaps();

GetPort(&gp);

/* clean-up current bitmaps */
if (WritePtr != myWindow) rmPix(WritePtr, mapptr);
if (PictPtr != 0) rmPix(PictPtr, pmapptr);
mapptr = pmapptr = 0;
textmap = 0;

/* save new pixel size */
GetGlobalRect(myWindow, &r);
bitmap_gdev = GetMaxDevice(&r);
currPM = (*bitmap_gdev)->gdPMap;
pxsize = (*currPM)->pixelSize;

/* restore settings before bitmap allocation */
SetPort(myWindow);
RGBBackColor(&realwhite);

/* Get color table updated for colors we want */
ActivatePalette(myWindow);

				/* attempt to allocate text bitmap */
WriteMap.bounds.top = 0;
WriteMap.bounds.left = 0;
WriteMap.bounds.bottom = vpixsize;
WriteMap.bounds.right = hpixsize;
mkPix(&WritePort, &mapptr, &(WriteMap.bounds), 0, bitmap_gdev);
textsize = pictsize;
pictsize = 0;
if (mapptr == 0) {
	WritePtr = myWindow;
	note_err(mapalrt);
	}
else {
	WritePtr = &WritePort;
	textmap = 1;
	}
SetPort(WritePtr);
GetEntryColor(myPalette, RGBback, &rgbtemp);
if (WritePtr != myWindow) {
	setgdev();
	RGBBackColor(&rgbtemp);
	resetgdev();
 	SetOrigin(textRect.left, textRect.top);
	ClipRect(&scrnRect);
	}
else RGBBackColor(&rgbtemp);
 								/* set-up graphics bit map */
pictRect.top = pictRect.left = 0;
pictRect.bottom = vpixsize-19;
pictRect.right = hpixsize;

PictPtr = &PictPort;
								/* set-up PictPort */
mkPix(&PictPort, &pmapptr, &pictRect, 0, bitmap_gdev);
if (pmapptr == 0) PictPtr = 0;
else {
	setgdev();
	PictPtr = &PictPort;
	GetEntryColor(myPalette, RGBback, &rgbtemp);
	RGBBackColor(&rgbtemp);
	resetgdev();
	}

SetPort(gp);
alignbitmaps();
invldscr();
newstat();
justGrowIcon(0);
}
 
macend()
{
static char * ScrDmpEnb = (char *)0x2f8;
scrapline * sp, * nextsp;


if (myWindow != 0) removewindow();

if (new_settings()) if (savedlg()) write_settings();

(*ScrDmpEnb) = ScrDmpSav;	/* restore original byte */

							/* clean-up printing */
if (prtinit) PrClose();
if (hPrint != 0) DisposHandle(hPrint);

							/* clean-up scrap */
removescrap();

							/* clean-up async sound */
if (sndactive) {
	SndDisposeChannel(scp, true);
	}
if (scp != 0) {
	DisposPtr(scp);
	}
if (sndhandle != 0) {
	DisposHandle(sndhandle);
	}
							/* clean-up API */
apideregister(0);

if (ppcavail) {
	endparamblocks();
	endclientrecs();
	enddatablocks();
	}
if (rcvMessage != 0) DisposHandle(rcvMessage);
if (sndMessage != 0) DisposHandle(sndMessage);
}						

hndmac()
{
	register short code;
	register unsigned short i, w, h;
	register long l;
	static unsigned long * Ticks = (unsigned long *)0x16a;
	static unsigned long * DoubleTime = (unsigned long *)0x2f0;
	static unsigned long clklimit = 0;
	static unsigned long limit = 0;
	unsigned long downticks;
	GrafPtr gp;
	Point pt;
	OSErr rc, alignbitmaps();
	char background, usenull;
	unsigned char asciicode, c, selflag; 
	unsigned char shiftflag, optionflag, commandflag, doubleflag;
	static short downloc = 0;
	static short uploc = 0;
	static short oldloc = 0;
	short moveloc, newloc;
	static unsigned long downtime = 0;
	static unsigned long uptime = 0;
	unsigned short j, k, hexcount;
	Size s;
	hexmap * hm;
	Point dipoint;

								/* I/O which must happen often */
	if (serflg) {
		hndser();
		serkbd();
		}
	else if (netconn) getreq();
	else if (tcpflg) tcpevent();
	
	if (sndactive == 2) {
		SndDisposeChannel(scp, true);
		sndactive = 0;
		}
	
								/* during FTP, take all we can get */
	background = mf_bgrnd && !(ftpcopen || ftpdopen || ftplopen);

	usenull = 1;				/* initially will do null processing */
								
	if (!background) {			/* don't do non-essential things too often */
								/* when running in the foreground */
		if (limit >= (*Ticks)) usenull = 0;
		else limit = (*Ticks) + 4;		/* set time for next null handling */
		}

						/* set sw_bgrnd to 2 if next GNE call will suspend us */
	if (sw_bgrnd == 1) sw_bgrnd = 2;
	if (background)
		 rc = WaitNextEvent(everyEvent, &myEvent, 4L, 0L);
	else rc = GetNextEvent(everyEvent, &myEvent);
	myStask();			/* needed for FTP in MultiFinder background */
	if ((rc == 0) && usenull)		/* FALSE from GNE */
		switch(myEvent.what) {
			case nullEvent:
					if (myWindow != 0) {
						if (clklimit < (*Ticks)) {		/* clock updating */
							clklimit = (*Ticks) + 600;
							newstat();
							}
						if (!mf_bgrnd) {
							/* handle cursor setting */
							if (FrontWindow() == myWindow) {
								if (((WindowPeek)myWindow)->visible) {
									GetPort(&gp);
									SetPort(myWindow);
									GetMouse(&pt);
									SetPort(gp);
									i = (PtInRgn(&pt, myWindow->visRgn) == 0);
									if (!i) i = (PtInRect(&pt, &gr_rect) != 0);
									if (dfltcurs != i) {
										if (i) SetCursor(&qd.arrow);
										else SetCursor(*grafcursor);
										dfltcurs = i;
										}
									}
								else {
									arrowcursor();
									}
								}
							/* check for color updating */
							if (colormac) {
								chkpx();
								if (newCTab()) {
									ActivatePalette(myWindow);
									if (updCTab()) {
										newbackcolor();
										invldscr();
										newstat();
										justGrowIcon(0);
										}
									}
								}
							}
						}
					if (!mf_bgrnd) {
						myStask();
						SystemTask();			/* run DAs, etc. */
						}
					if (ppcavail) {			/* check if allocations needed */
						addparamblocks();
						addclientrecs();
						adddatablocks();
						}
					apinewmsg();			/* check for API event */
					break;
			default:
					return;
			}
	else if (rc != 0) switch(myEvent.what) {		/* true from GNE */
		case mouseDown:
			code = FindWindow(&myEvent.where, &whichWindow);
			switch (code) {
			case inMenuBar:
				menu_upd();
				docommand(MenuSelect(&myEvent.where),
						  (myEvent.modifiers & shiftKey) != 0);
				break;
			case inSysWindow:
				SystemClick(&myEvent, whichWindow);
				break;
			case inDrag:
				if (whichWindow == FrontWindow()) {
					DragWindow(whichWindow, &myEvent.where,
								&dragRect);
					if ((whichWindow == myWindow) && (myWindow != 0)) {
						rc = alignbitmaps();
						if (rc == 0) {
							invldscr();
							newstat();
							}
						}
					}
				else {
					SelectWindow(whichWindow);
					}
				break;
			case inGoAway:
				if (xdlg && (whichWindow == xdlgptr)) {
					if (TrackGoAway(whichWindow, &myEvent.where))
						kabort = 1;
					break;
					}
				if ((whichWindow == myWindow) && (myWindow != 0))
					if (TrackGoAway(whichWindow,&myEvent.where))
						if (wants_to_quit()) {
							if (serflg) {
								if (connflg) {
									done = 1;
									}
								else {
									closeresponse(closeUser);
									if (logon) serlgout(0xe3);
									else serlgout(0xe7);
									}
								}
							else {
								closeresponse(closeUser);
								if (tcpflg) {
									tcplgout();
									}
								else {
									netlgout();
									}
								}
							}
				break;
			case inGrow:
				if (whichWindow == FrontWindow()) {
					GetPort(&gp);
					SetPort(whichWindow);
					if ((whichWindow == myWindow) && (myWindow != 0)) {
						SetClip(cliprgn);
						}
					l = GrowWindow(whichWindow, &myEvent.where,
								&sizeRect);
					h = l >> 16;
					w = l & 0x0000ffffL;
					SizeWindow(whichWindow, w, h, true);
					if ((whichWindow == myWindow) && (myWindow != 0)) {
						InvalRect(&gr_rect);
						InitClip();
						InvalRect(&gr_rect);
						}
					if (xdlg && (whichWindow == xdlgptr)) {
						EraseRect(&(whichWindow->port.portRect));
						InvalRect(&(whichWindow->port.portRect));
						}
					SetPort(gp);
					}
				break;
			case inContent:
				if (whichWindow != FrontWindow()) {
					SelectWindow(whichWindow);
					break;
					}
				if ((whichWindow == myWindow) && (myWindow != 0)) {
					downtime = myEvent.when;
					shiftflag = (myEvent.modifiers & shiftKey) != 0;
					optionflag = (myEvent.modifiers & optionKey) != 0;
					commandflag = (myEvent.modifiers & cmdKey) != 0;
					GetPort(&gp);
					SetPort(myWindow);
					GlobalToLocal(&myEvent.where);
					
					/* calculate location for cursor movement (moveloc) */
					x = (myEvent.where.h-2+htxtstrt)/htxtsize;
					y = myEvent.where.v/vtxtsize;
					if (x < 0) x = 0;
					if (x > scrhoff) x = scrhoff;
					if (y < 0) y = 0;
					if (y > scrvoff) y = scrvoff;
					moveloc = (y*scrhsize) + x;

					/* check if mouse down is close enough in time and 	*/
					/* location to be a valid double click 				*/
					doubleflag = ((downtime - uptime) <= (*DoubleTime)) &&
								 (moveloc == uploc);
					uptime = 0;
								
					/* calculate location for selection (downloc) */
					x = (myEvent.where.h-2+htxtstrt)/htxtsize;
					y = myEvent.where.v/vtxtsize;
					if (x < 0) x = 0;
					if (x > scrhoff) {
						if (optionflag) x = scrhoff;
						else x = scrhsize;
						}
					if (y < 0) y = 0;
					if (y > scrvoff) {
						if (optionflag) downloc = (scrvoff*scrhsize) + x;
						else downloc = maxcnt;
						}
					else downloc = (y*scrhsize) + x;
					
					newloc = downloc;
					selflag = 0;
					downtime = *Ticks;
					while (StillDown()) {
						GetMouse(&pt);
						x = (pt.h-2+htxtstrt)/htxtsize;
						y = pt.v/vtxtsize;
						if (x < 0) x = 0;
						if (x > scrhoff) {
							if (optionflag) x = scrhoff;
							else x = scrhsize;
							}
						if (y < 0) y = 0;
						if (y > scrvoff) {
							if (optionflag) newloc = (scrvoff*scrhsize) + x;
							else newloc = maxcnt;
							}
						else newloc = (y*scrhsize) + x;
						if ((selflag == 0) && (newloc != downloc)) { /* sel init */
							if (optionflag == 0) {
								if (shiftflag) {
									checksel();
									if (textsel) addsel(&downloc); /* reset downloc */
									}
								arrowcursor();
								}
							selflag = 1;
							mousesel(0, downloc, doubleflag);
							}
						if (selflag) {
							if (optionflag) mousesel(1, newloc, doubleflag);	
							else mousesel(2, newloc, doubleflag);
							}
						if (tcpflg) myStask();
						}
					downtime = (*Ticks) - downtime;
					SetPort(gp);
					if (selflag) mousesel(3, 0, doubleflag);	/* termination */
					if (selflag && 		/* don't use singe position selection */
						((newloc - downloc > 1) || 
						 (newloc - downloc < -1) ||
						 (downtime > 60) ||
						 doubleflag )) {
						if (newloc != downloc) {
							if (optionflag) {
								if (!shiftflag) resetsel();
								setrectsel(newloc, downloc);
								if (doubleflag) addfieldsel();
								}
							else {
								if (commandflag == 0) resetsel();
								if (doubleflag) wordadj(&downloc, &newloc, 1);
								settextsel(newloc, downloc, commandflag);
								}
							}
						if (serflg && connflg) break;
						if (tcpflg && (!logon)) break;
						if (doubleflag) {
							curadr = oldloc;
							newcur();
							}
						break;
						}
					if ((atrbuff[downloc] & 0x0020) && (!shiftflag)
						&& (!doubleflag)) {
						resetsel();
						break;
						}
					if (serflg && connflg) break;
					if (tcpflg && (!logon)) break;
					oldloc = curadr;
					curadr = moveloc;
					newcur();
					if (doubleflag) {
						if (kblock || (vmxbgn && (!vmxsub))) beep();
						else clickattn(cs.mousepf);
						}
					}
				break;

			case inZoomIn:
			case inZoomOut:
				if (TrackBox(whichWindow, &myEvent.where, code))
					dozoom(whichWindow, code);
				break;

			default:
				break;
			}
			break;

		case mouseUp:
			if (FindWindow(&myEvent.where, &whichWindow) !=
				inContent) break;
			if (myWindow == 0) break;
			if (whichWindow != myWindow) break;
			if (myWindow != FrontWindow()) break;
			uptime = myEvent.when;
			GetPort(&gp);
			SetPort(myWindow);
			GlobalToLocal(&myEvent.where);
			SetPort(gp);
			x = (myEvent.where.h-2+htxtstrt)/htxtsize;
			y = myEvent.where.v/vtxtsize;
			if (x < 0) x = 0;
			if (x > scrhoff) x = scrhoff;
			if (y < 0) y = 0;
			if (y > scrvoff) y = scrvoff;
			uploc = (y*scrhsize) + x;
			break;

		case keyDown:
										/* check for special user */
			if ((myEvent.modifiers & 0x0f00) == 0x0f00) {
				if (((myEvent.message & 0x7f00) >> 8) == 0x24) {
					specuser ^= 1;
					break;
					}
				}
		case autoKey:
			if (apiopen) {
				if (apiopenreq.hide) {
					apiopenreq.hide = 0;	/* to allow beep to work */
					beep();
					apiopenreq.hide = 1;
					}
				}
			asciicode = 0xff;
			if ((connflg && serflg) ||
				(tcpflg && (pend_conn == 3))) {
							/* line mode connection */
				i = myEvent.message & 0x7f;
				if (i == 0x60) i = 0x1b;
				if (myEvent.modifiers & cmdKey) {
					if (i == 0x1b) i = 0x60;
					else {
						if (i == 0x08) i = 0x7f;
						else i &= 0x1f;
						}
					}
				c = i;
				if (serflg) {
					ObscureCursor();
					if (c == 0x12) {
						serverflags = 0x01;
						srvswitch();
						}
					else {
						if (servermode) {
							putsrv(&c, 1, 1);
							}
						else {
							mout(c);
							}
						}
					break;
					}
						/* for tcp, echo unless telnet host is doing it */
				if (servermode) {
					putsrv(&c, 1, 1);
					break;
					}
				if (!hisopts[TELOPT_ECHO])
					putscr(&c, 1, 0);
				asciicode = i;
				}
			i = (myEvent.message & 0x7f00) >> 8; /* key code */
			
								/* check special serial mode control characters */
			if (serflg && (myEvent.modifiers & cmdKey)) {
				if (i == 6) {
					vmkio(255, 0);	/* cmd-Z */
					break;
					}
				}
			
									/* apply modifiers for keycode test */
			if ((myEvent.modifiers & (cmdKey+shiftKey)) ==
				(cmdKey+shiftKey)) i += 384;
			else if (myEvent.modifiers & cmdKey) i += 256;
			else if (myEvent.modifiers & optionKey) i += 512;
			else if (myEvent.modifiers & shiftKey) i += 128;
									/* get array entry */
			k = (*kb_handle)[i];
									/* if no keycode entry, try ASCII code */
			if (k == 0) {
				i = myEvent.message & 0xff;			/* ASCII code */
				if ((myEvent.modifiers & (cmdKey+shiftKey)) ==
					(cmdKey+shiftKey)) i += 768;
				else if (myEvent.modifiers & cmdKey) i += 512;
				else if (myEvent.modifiers & optionKey) i += 1024;
				else if (myEvent.modifiers & shiftKey) i += 256;
				i += 640;			/* skip over keycode tables */
									/* get array entry */
				k = (*kb_handle)[i];
				}
				
									/* if no keycode, check for menu equivalent */
			if (k == 0) {
				if ((myEvent.modifiers & cmdKey) && (myEvent.what == keyDown)) {
					menu_upd();
					docommand(MenuKey(myEvent.message & 0xff),
						  (myEvent.modifiers & shiftKey) != 0);
					}
				break;
				}
			k -= 1;
			if (myEvent.what == autoKey)
				if ((kbtyp[k] & 0x0f) == 2) break;
			
			/* check for overriding hex mapping */
			if (k == 254) {
				/* no typeahead for special hex chars */
				if (kbqsize > 0) {
					beep();
					break;
					}
				s = GetHandleSize(kb_handle);
				if (s > 1920) {
					hexcount = (s - 1920)/4;
					hm = (hexmap *)((*kb_handle) + 1920);
					kbstd[k] = 0;
					for (j=0; j < hexcount; j++) {
						if (hm->offset == i) {
							kbstd[k] = hm->stdchar;
							kbapl[k] = hm->aplchar;
							break;
							}
						hm++;
						}
					if (kbstd[k] == 0) break;
					}
				else break;
				}
			
			ObscureCursor();
			h = (myEvent.modifiers & alphaLock) != 0;
			if (serflg) vmkio(k, h);
			else if (tcpflg) tcpin(asciicode, k, h);
			else netin(k, h);
			break;
 
		case activateEvt:
			dfltcurs = 2;			/* force cursor to be set */
			if (myEvent.modifiers & activeFlag) {	/* activate event */
				if (da_menu != 0) {
					da_menu = 0;
					SetItem(myMenus[1], 19, "Quit");
					}
				}
			else {									/* deactivate event */
				if (da_menu == 0) {
					da_menu = 1;
					SetItem(myMenus[1], 19, "Close");
					}
				if (((struct GrafPort *)myEvent.message == myWindow)
					 && (myWindow != 0)) {
						arrowcursor();
						}
				}
			appl_menu();
			if (((struct GrafPort *)myEvent.message == myWindow) &&
				(myWindow != 0)) {
				if (myEvent.modifiers & activeFlag) {	/* activate event */
					if (colormac) {
						updCTab();
						newbackcolor();
						invldscr();
						cs.stat_time += 8;
						newstat();
						chkpx();
						}
					}
				justGrowIcon(0);		/* activate or deactivate */
				}
			else if (((struct GrafPort *)myEvent.message == xdlgptr)
				&& xdlg) xfGrowIcon();
			break;
 
		case updateEvt:
			updevent(myEvent.message);
			break;

		case app4Evt:		/* probably switcher */
			if ((myEvent.message & 0xff000000) != 0x01000000)
				break;	/* not switcher event */
			else if (myEvent.message & 0x00000001) sw_resume(myEvent.message);
					else sw_suspend(myEvent.message);
			break;

		case diskEvt:	/* if mount failed, put up disk init dialog */
			if ((myEvent.message & 0xffff0000) != 0) {
				dipoint.h = kDILeft;
				dipoint.v = kDITop;
				DIBadMount(&dipoint, myEvent.message);
				}
			break;

		default:	break;
		}
}

dozoom(wp, code)
struct WindowRecord * wp;
short code;
{
GrafPtr gp;
WStateData * wsptr;
OSErr rc;

						/* be sure this window can zoom */
if (((WindowPeek)wp)->dataHandle == 0) return;

GetPort(&gp);
SetPort(wp);
if ((wp == myWindow) && (myWindow != 0)) SetClip(cliprgn);
EraseRect(&(wp->port.portRect));
ZoomWindow(wp, code, false);
if ((wp == myWindow) && (myWindow != 0)) {
	InitClip();
	InvalRect(&gr_rect);
	rc = alignbitmaps();
	if (rc == 0) {
		invldscr();
		newstat();
		}
	}
SetPort(gp);
}

short zoomstate()
{
struct Point topleft, bottomright;
Rect r;
WStateData * wsptr;
GrafPtr gp;

if (myWindow == 0) return(0);	/* return if no window */
								/* be sure my window can zoom */
if (((WindowPeek)myWindow)->dataHandle == 0) return(0);

								/* convert portRect to global coordinates */
GetGlobalRect(myWindow, &r);
								/* check for standard vs. user state */
wsptr = (struct WStateData *)(*(((WindowPeek)myWindow)->dataHandle));
if ((wsptr->stdState) == r) return(inZoomIn);	/* return code */
else return(inZoomOut);
}

chkpx() {				/* handle user changing pixel bit depth */
GDHandle currGD;
PixMapHandle currPM, myPM, newPM;
GrafPort gp, newport;
char * newptr;
RGBColor rgbtemp;
Rect r;

if (!colormac) return;
GetGlobalRect(myWindow, &r);
currGD = GetMaxDevice(&r);
currPM = (*currGD)->gdPMap;
if (pxsize == (*currPM)->pixelSize) return;
newdepth();
}

fixcolors(p)		/* fixcolors is called to handle a redefinition of		*/
PaletteHandle p;	/* one of the "usr" colors.  It updates the specified	*/
					/* palette and associated window.						*/
{
short i, usage;
RGBColor rgbtemp;

if (cs.exactcolor) {
	SetEntryColor(p, RGBblack, &cs.usrblack);
	SetEntryColor(p, RGBwhite, &cs.usrwhite);
	SetEntryColor(p, RGBback, &cs.usrback);
	SetEntryColor(p, RGBgreen, &cs.usrgreen);
	SetEntryColor(p, RGBred, &cs.usrred);
	SetEntryColor(p, RGBblue, &cs.usrblue);
	SetEntryColor(p, RGBpink, &cs.usrpink);
	SetEntryColor(p, RGByellow, &cs.usryellow);
	SetEntryColor(p, RGBturquoise, &cs.usrturquoise);
	SetEntryColor(p, RGBstat, &cs.usrstat);
	}
else {
	mapcolor(&cs.usrblack, &rgbtemp);
	SetEntryColor(p, RGBblack, &rgbtemp);
	SetEntryColor(p, RGBwhite, &cs.usrwhite);
	SetEntryColor(p, RGBback, &cs.usrback);
	mapcolor(&cs.usrgreen, &rgbtemp);
	SetEntryColor(p, RGBgreen, &rgbtemp);
	mapcolor(&cs.usrred, &rgbtemp);
	SetEntryColor(p, RGBred, &rgbtemp);
	mapcolor(&cs.usrblue, &rgbtemp);
	SetEntryColor(p, RGBblue, &rgbtemp);
	mapcolor(&cs.usrpink, &rgbtemp);
	SetEntryColor(p, RGBpink, &rgbtemp);
	mapcolor(&cs.usryellow, &rgbtemp);
	SetEntryColor(p, RGByellow, &rgbtemp);
	mapcolor(&cs.usrturquoise, &rgbtemp);
	SetEntryColor(p, RGBturquoise, &rgbtemp);
	mapcolor(&cs.usrstat, &rgbtemp);
	SetEntryColor(p, RGBstat, &rgbtemp);
	}

if (cs.exactcolor) usage = pmTolerant;
		   else usage = pmCourteous;
for (i=1; i < 10; i++)
	SetEntryUsage(p, i, usage, 0);
}

mapcolor(in, out)
RGBColor * in, * out;
{
long l;
short i;
RGBColor csave;

if (textmap) setgdev();
l = Color2Index(in);
i = l &0x0000ffff;
csave = *in;
if (i == 0) {
	csave = *in;
	InvertColor(&csave);
	*out = csave;
	}
else *out = *in;
if (textmap) resetgdev();
}

actualcolor(in, out)  /* convert color to actual color displayed */
RGBColor * in, *out;
{
if (textmap) setgdev();
Index2Color(Color2Index(in), out);
if (textmap) resetgdev();
}

updevent(msgptr)
GrafPtr msgptr;
{
short i, txtmode;
long t;
GrafPtr gp;
BitMap * srcbits;
char statflg;
PixMapHandle currPM;

statflg = 0;
GetPort(&gp);
SetPort(msgptr);
if ((msgptr == myWindow) && (myWindow != 0)) {
	SetClip(cliprgn);
	}
BeginUpdate(msgptr);
if ((msgptr == myWindow) && (myWindow != 0)) {
	if (WritePtr != myWindow) {
		if (RectInRgn(&textRect, myWindow->visRgn) && needwrite) {
			SetPort(WritePtr);
			setgdev();
			writescr(0);
			resetgdev();
			SetPort(myWindow);
			needwrite = 0;
			if (wr_active) statflg = 1;
			}
		if (colormac) {
			RGBForeColor(&realblack);
			RGBBackColor(&realwhite);
			}
		if (drawpict && (PictPtr != 0)) {
			if (colormac)
				srcbits = (BitMap *)(*(PixMapHandle)PictPort.portBits.baseAddr);
			else
				srcbits = &PictPort.portBits;
			CopyBits(srcbits, &(myWindow->portBits),
							&pictRect, &textRect, srcCopy, 0L);
			}
		if (colormac)
			srcbits = (BitMap *)(*(PixMapHandle)WritePort.portBits.baseAddr);
		else
			srcbits = &WritePort.portBits;
		if (drawpict) {
			if (colormac) {
				currPM = (*bitmap_gdev)->gdPMap;
				if (((*currPM)->pixelSize) > 1) {
					if (cs.invertbw) txtmode = adMin;
					else txtmode = adMax;
					}
				else {
					if (cs.invertbw) txtmode = srcOr;
					else txtmode = notSrcBic;
					}
				}
			else {
				txtmode = srcOr;
				}
			CopyBits(srcbits, &(myWindow->portBits),
		 			&textRect, &textRect, txtmode, 0L);
			CopyBits(srcbits, &(myWindow->portBits),
		 			&statRect, &statRect, srcCopy, 0L);
			}
		else CopyBits(srcbits, &(myWindow->portBits),
		 		&scrnRect, &scrnRect, srcCopy, 0L);
		drawcurs(0);		/* draw cursor, no restore */
		if (RectInRgn(&gr_rect, myWindow->visRgn)) justGrowIcon(1);
		}
	else {
		if (RectInRgn(&textRect, myWindow->visRgn)) {
			writescr(0);
			needwrite = 0;
			if (wr_active) statflg = 1;
			}
		if (RectInRgn(&statRect, myWindow->visRgn)) {
			statline();
			}
		if (RectInRgn(&gr_rect, myWindow->visRgn)) justGrowIcon(1);
		}
	}
else if ((xdlg == 1) && (msgptr == xdlgptr)) {	
		DrawDialog(xdlgptr);
		xfGrowIcon();
		}
EndUpdate(msgptr);
if ((msgptr == myWindow) && (myWindow != 0)) SetClip(destrgn);
SetPort(gp);
if (statflg) {
	wr_active = 0;
	newstat();
	}
if (showstg) {
	showstg = 0;
	stginfo();
	}
}

mkPix(gptr, mptr, b_rect, mustalloc, gdev)
GrafPtr gptr;
char ** mptr;
Rect * b_rect;
char mustalloc;
GDHandle gdev;
{
	GDHandle saveGD;
	Size mmapsize;
	Ptr temp;
	PixMapPtr pixptr;
	PixMapHandle devPM;
	CTabHandle myCTabHandle;
	OSErr memerr;

	saveGD = GetGDevice();
	SetGDevice(gdev);
	OpenCPort(gptr);
	pixptr = *(PixMapHandle)(gptr->portBits.baseAddr);
	pixptr->bounds = *b_rect;
	PortSize(pixptr->bounds.right, pixptr->bounds.bottom);
	SetRectRgn(gptr->visRgn, 0, 0, pixptr->bounds.right,
				pixptr->bounds.bottom);
	devPM = (*gdev)->gdPMap;
	pixptr->rowBytes = ((((hpixsize * (*devPM)->pixelSize) + 31) >> 5) + 1) << 2;
	mmapsize = pixptr->bounds.bottom - pixptr->bounds.top;
	mmapsize = mmapsize * pixptr->rowBytes + 4;
	pictsize = mmapsize;
	*mptr = myNewPtr(mmapsize);
	temp = *mptr;
	while (((Size)temp%4) != 0) temp++;	/* align on long word */
	pixptr = *(PixMapHandle)(gptr->portBits.baseAddr);
	pixptr->baseAddr = temp;
	if (pixptr->baseAddr == 0L) {		/* no bitmap */ 
		if (mustalloc) {
			SetGDevice(saveGD);
			stgalert("color bitmap", "NewPtr", mmapsize);
			ExitToShell();
			}
		else {
			SetGDevice(saveGD);
			CloseCPort(gptr);
			return;
			}
		}
	pixptr->rowBytes += 0x8000;

	/* set-up color table for our map */
	devPM = (*gdev)->gdPMap;
	myCTabHandle = (*devPM)->pmTable;
	memerr = HandToHand(&myCTabHandle);
	if (memerr != 0)
		if (mustalloc) {
			SetGDevice(saveGD);
			stgalert("color table", "NewPtr", mmapsize);
			ExitToShell();
			}
		else {
			SetGDevice(saveGD);
			DisposPtr(*mptr);
			*mptr = 0;
			CloseCPort(gptr);
			return;
			}
	pixptr = *(PixMapHandle)(gptr->portBits.baseAddr);
	pixptr->pmTable = myCTabHandle;

	/* miscellaneous stuff */
	
	ClipRect(&(gptr->portRect));
	EraseRect(&(gptr->portRect));
	SetOrigin(0, 0);
	TextFont(stdfont);
	TextSize(cur_ptsize);
	SetGDevice(saveGD);
}

Ptr myNewPtr(s)
Size s;
{
Ptr result;
long l, reserve;
long tcpgetrsrv();

result = NewPtr(s);
if (result == 0) return(result);

/* check there is still enough free storage for normal operation */
MaxApplZone();
l = FreeMem();
reserve = 32768;		/* need at least 32K in any case */
if (tcpflg) {				/* adjust for TCP/IP */
	reserve += tcpgetrsrv(0);	/* was tcpgetrsrv(initflg) */
	}
else {						/* adjust for serial and appletalk */
/*	if (initflg) reserve += 40960;	*/
	}
if (l < reserve) {
	DisposPtr(result);
	result = 0;
	}
return(result);
}

int updCTab()				/* check for new color table */
{
GDHandle currGD;
PixMapHandle currPM, myPM;
CTabHandle currcth, mycth;
OSErr memerr;
Rect r;

if (!textmap) return(0);

GetGlobalRect(myWindow, &r);
currGD = GetMaxDevice(&r);
currPM = (*currGD)->gdPMap;
currcth = (*currPM)->pmTable;

myPM = (PixMapHandle)(WritePtr->portBits.baseAddr);
mycth = (*myPM)->pmTable;

/* return if color table hasn't changed */
if ((*currcth)->ctSeed == (*mycth)->ctSeed) return(0);

/* return if pixel depth has changed */
if ((*myPM)->pixelSize != (*currPM)->pixelSize) return(0);

/* make my color table match the new one */
DisposHandle(mycth);

currGD = GetMaxDevice(&r);
currPM = (*currGD)->gdPMap;
currcth = (*currPM)->pmTable;
myPM = (PixMapHandle)(WritePtr->portBits.baseAddr);

(*myPM)->pmTable = currcth;
memerr = HandToHand(&((*myPM)->pmTable));
if (memerr != 0) {
	stgalert("text color table", "HandToHand", 0L);
	ExitToShell();
	}

/* repeat for graphics bitmap */
if (PictPtr == 0) return(1);
myPM = (PixMapHandle)(PictPtr->portBits.baseAddr);
mycth = (*myPM)->pmTable;
/* make my color table match the new one */
DisposHandle(mycth);

currGD = GetMaxDevice(&r);
currPM = (*currGD)->gdPMap;
currcth = (*currPM)->pmTable;
myPM = (PixMapHandle)(PictPtr->portBits.baseAddr);

(*myPM)->pmTable = currcth;
memerr = HandToHand(&((*myPM)->pmTable));
if (memerr != 0) {
	stgalert("graphics color table", "HandToHand", 0L);
	ExitToShell();
	}
return(2);
}

int newCTab()				/* check for new color table */
{
GDHandle currGD;
PixMapHandle currPM, myPM;
CTabHandle currcth, mycth;
Rect r;

if (!textmap) return(0);

GetGlobalRect(myWindow, &r);
currGD = GetMaxDevice(&r);
currPM = (*currGD)->gdPMap;
currcth = (*currPM)->pmTable;

myPM = (PixMapHandle)(WritePtr->portBits.baseAddr);
mycth = (*myPM)->pmTable;

/* return 0 if color table hasn't changed */
if ((*currcth)->ctSeed == (*mycth)->ctSeed) return(0);

/* return 0 if pixel depth has changed */
if ((*myPM)->pixelSize != (*currPM)->pixelSize) return(0);

/* else return 1 */
return(1);
}

rmPix(gptr, mptr)
GrafPtr gptr;
Ptr mptr;
{
PixMapPtr pixptr;

									/* dispose copy of color table */
pixptr = *(PixMapHandle)(gptr->portBits.baseAddr);
DisposHandle(pixptr->pmTable);
if (mptr != 0) DisposPtr(mptr);
pixptr->baseAddr = 0;
CloseCPort(gptr);
}

newinvbw()		/* make updates to handle a new value of invbw */
{
if (!colormac) return;

if (cs.invertbw) {
	if (curnum > 20846) curnum -= 2;
	else return;
	}
else {
	if (curnum < 20847) curnum += 2;
	else return;
	}
grafcursor = GetCursor(curnum);
dfltcurs = 2;		/* force SetCursor call */
}

newbackcolor()
{
RGBColor rgbtemp;
GrafPtr gp;

GetEntryColor(myPalette, RGBback, &rgbtemp);
GetPort(&gp);
SetPort(WritePtr);
if (textmap) setgdev();
RGBBackColor(&rgbtemp);
if (PictPtr != 0) {
	SetPort(PictPtr);
	RGBBackColor(&rgbtemp);
	}
SetPort(gp);
if (textmap) resetgdev();
}

setgdev()
{
if (!colormac) return;
saved_gdev = GetGDevice();
SetGDevice(bitmap_gdev);
}

resetgdev()
{
if (!colormac) return;
SetGDevice(saved_gdev);
}

adjfmt(changed, windmax, dfltptsize, altptsize, altrows, altcols)
char *changed, *windmax;
short *dfltptsize, *altptsize, *altrows, *altcols;
/* adjust screen format settings to be valid for the current set of gdevs */
{
char altok;
short altsize, newptsize, newrows, newcols, maxrows, maxcols;

(*changed) = 0;

/* set default point size to 9 if 24-by-80 won't fit with 12-point */
if ((rowmax12 < 24) || (colmax12 < 80)) {
	if ((*dfltptsize) == 12) {
		(*dfltptsize) = 9;
		(*changed) = 1;
		}
	}
/* check alternate screen size is valid */
if ((*altrows) < 24) {
	(*altrows) = 24;
	(*changed) = 1;
	}
if ((*altcols) < 80) {
	(*altcols) = 80;
	(*changed) = 1;
	}

/* check if alternate screen size will fit */
if ((*altptsize) == 9) {
	altok = (((*altrows) <= rowmax9) && ((*altcols) <= colmax9));
	}
else {
	altok = (((*altrows) <= rowmax12) && ((*altcols) <= colmax12));
	}
/* if not, and 12-point specified, try 9-point */
if (((*altptsize) == 12) && (!altok)) {
	altok = (((*altrows) <= rowmax9) && ((*altcols) <= colmax9));
	if (altok) {
		(*altptsize) = 9;	
		(*changed) = 1;
		}
	}
if (altok) {
	return;
	}
	
/* find the closest standard model that will fit, and use the largest point
   size for that model */

altsize = (*altrows) * (*altcols);
if ((altsize >= 3564) && ((*altcols) > 80)) {
	newrows = 27;	/* model 5 */
	newcols = 132;
	}
else if (altsize >= 3440) {
	newrows = 43;	/* model 4 */
	newcols = 80;
	}
else if (altsize >= 2560) {
	newrows = 32;	/* model 3 */
	newcols = 80;
	}
else {
	newrows = 24;	/* model 2 */
	newcols = 80;
	}
newptsize = 12;
maxrows = rowmax12;
maxcols = colmax12;
while ((newrows > 24) || (newptsize > 9)) {
	/* test this model/ptsize */
	if ((newrows <= maxrows) && (newcols <= maxcols)) break; 
	/* set next model/ptsize */
	if (newptsize == 12) {
		newptsize = 9;
		maxrows = rowmax9;
		maxcols = colmax9;
		}
	else {
		newptsize = 12;
		maxrows = rowmax12;
		maxcols = colmax12;
		newcols = 80;
		if (newrows == 27) newrows = 43;
		else if (newrows == 43) newrows = 32;
		else if (newrows == 32) newrows = 24;
		}
	}
(*altptsize) = newptsize;
(*altrows) = newrows;
(*altcols) = newcols;
(*changed) = 1;
}

newstdfont(f)
short f;
{
GrafPtr gp;

stdfont = f;
fixbracket = cs.std_brack && (!aplmode) && (stdfont != ALAFONT);
if (myWindow == 0) return;
GetPort(&gp);
SetPort(myWindow);
TextFont(f);
if (WritePtr != 0) {
	SetPort(WritePtr);
	TextFont(f);
	}
if (PictPtr != 0) {
	SetPort(PictPtr);
	TextFont(f);
	}
SetPort(gp);
}

updwindpos()
{
Point p;
GrafPtr gp;

if (myWindow == 0) return;

p.h = (myWindow->portRect.right + myWindow->portRect.left + 1)/2;
p.v = (myWindow->portRect.bottom + myWindow->portRect.top - 17)/2;
GetPort(&gp);
SetPort(myWindow);
LocalToGlobal(&p);		/* global current center */
SetPort(gp);
cs.windpth = p.h;
cs.windptv = p.v;
wposok = 1;
}
